#include <QApplication>
#include <QLabel>
#include <QVBoxLayout>
#include <QStyle>

#include <QMoveEvent>
#include <QWindow>
#include <QScreen>

#include <QDebug>
#include <QSysInfo>
#include <QTimer>

#include <algorithm>
#include <iostream>

// 1.2

bool optWordWrap = true;
int n = 0;

static QString title()
{
    return QLatin1String(qVersion()) + QLatin1Char(' ') + QGuiApplication::platformName()
        + QLatin1Char(' ') + QApplication::style()->objectName();
}

struct formatRect
{
    formatRect(const QRect &r) : m_r(r) {}

    QRect m_r;
};

QDebug operator<<(QDebug d, const formatRect &r)
{
    QDebugStateSaver s(d);
    d.nospace();
    d << r.m_r.width() << 'x' << r.m_r.height() << Qt::forcesign
      << r.m_r.left() << r.m_r.top() << Qt::noforcesign;
    return d;
}

struct formatWidgetGeom
{
    formatWidgetGeom(const QWidget *w) : m_w(w) {}

    const QWidget *m_w;
};

template <class Str>
static void formatScreen(const QScreen *s, Str & stream)
{
    auto g = s->geometry();
    stream << s->name() << "\", " << formatRect(g) <<  ", dpr=" << s->devicePixelRatio()
         << ", " << s->logicalDotsPerInchX() << "DPI";
}

static const int interval = 1000;

QDebug operator<<(QDebug d, const formatWidgetGeom &g)
{
    QDebugStateSaver s(d);
    d.nospace();
    d.noquote();
    QString on = g.m_w->objectName();
    if (!on.isEmpty())
        d << '"' <<on << "\"/";
    auto geom = g.m_w->geometry();
    auto frameGeom = g.m_w->frameGeometry();
    d << g.m_w->metaObject()->className() << ' '
      << formatRect(geom) << ' ';
    if (geom != frameGeom) {
        d << "fr=" << formatRect(frameGeom) << " fr[" << (geom.left() - frameGeom.left())
          << ',' << (geom.top() - frameGeom.top()) << ']';
    } else {
        d << "frameless";
    }
    if (auto w = g.m_w->windowHandle()) {
        if (w->handle())
        d << " wid=" << Qt::hex << Qt::showbase << w->winId() << Qt::dec << Qt::noshowbase;
    }
    return d;
}

class EventFilter : public QObject {
public:
    using QObject::QObject;

    bool eventFilter(QObject *, QEvent *) override;
};

bool EventFilter::eventFilter(QObject *o, QEvent *e)
{
    if (!o->isWidgetType())
        return false;
    auto w = static_cast<QWidget *>(o);
    if (!w->isWindow())
        return false;
    switch (e->type()) {
    case QEvent::Resize: {
        auto r = static_cast<QResizeEvent *>(e);
        qDebug() << o->metaObject()->className() << "ResizeEvent" << r->oldSize()
                 << r->size();
    }
        break;
    case QEvent::Move: {
        auto m = static_cast<QMoveEvent *>(e);
        qDebug() << w->metaObject()->className() << "MoveEvent"
                 << m->oldPos() << m->pos();
    }
        break;
    default:
        break;
    }
    return false;
}

class Widget  : public QWidget {
    Q_OBJECT
public:
    Widget();

public slots:
    void screenChanged();
};


Widget::Widget()
{
    QLabel *label = new QLabel;

    qDebug() << "optWordWrap=" << optWordWrap;
    label->setWordWrap(optWordWrap);
    label->setText("The main thread can start new threads by creating objects of a QThread subclass, as we did in the previous section. If these new threads need to communicate among themselves, they can use shared variables together with mutexes, read-write locks, semaphores, or wait conditions. But none of these techniques can be used to communicate with the main thread, since they would lock the event loop and freeze the user interface.");

    QVBoxLayout *layout = new QVBoxLayout(this);
    layout->addWidget(label);
}

void Widget::screenChanged()
{
    qDebug().noquote() << __FUNCTION__ << screen()->name();
}

int main(int argc, char *argv[])
{
    QStringList args;
    std::copy(argv + 1, argv + argc, std::back_inserter(args));

    QCoreApplication::setAttribute(Qt::AA_UseHighDpiPixmaps);
    QCoreApplication::setAttribute(Qt::AA_EnableHighDpiScaling);

#if QT_VERSION >= 0x050E00
    auto rp = Qt::HighDpiScaleFactorRoundingPolicy::PassThrough; // with or without this line doesn't make any difference
    int roundIndex = args.indexOf("-r");
    if (roundIndex >= 0 && roundIndex < args.size() - 1) {
        ++roundIndex;
        bool ok;
        int rpi = args.at(roundIndex).toInt(&ok);
        if (!ok || rpi  < 0 || rpi > int(Qt::HighDpiScaleFactorRoundingPolicy::PassThrough)) {
            qWarning() << "Invalid RoundingPolicy" << args.at(roundIndex);
            return -1;
        }
        rp = static_cast<Qt::HighDpiScaleFactorRoundingPolicy>(rpi);
    }
#endif // Qt 5.14
    qDebug() << rp;
    QGuiApplication::setHighDpiScaleFactorRoundingPolicy(rp);

    QApplication app(argc, argv);
    if (!args.contains("-e"))
        app.installEventFilter(new EventFilter(&app));

    QString t = title();

    {
        auto dbg = qDebug().noquote().nospace();
        dbg << t << args << '\n';
#if QT_VERSION >= 0x050E00
        dbg << "highDpiScaleFactorRoundingPolicy=" << QGuiApplication::highDpiScaleFactorRoundingPolicy() << '\n';
#endif // Qt 5.14
        for (auto s :  QGuiApplication::screens()) {
            formatScreen(s, dbg);
            dbg << '\n';
        }
    }

    optWordWrap = !args.contains("-w");

    Widget w;
    w.setWindowTitle(t);

    if (args.contains("f")) {
        qDebug("resize");
        w.resize(500, 300);  // NOTE: change `setFixedSize` to `resize` is a workaround, but... I need a fixedsize window
    } else {
        qDebug("fixed");
        w.setFixedSize(500, 300);
    }

    w.show();
    QObject::connect(w.windowHandle(), &QWindow::screenChanged, &w, &Widget::screenChanged);

    return QCoreApplication::exec();
}

#include "main.moc"
